#------------------------------------------------------------------
# $Copyright: Copyright (c) 2020 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/JBoss/JBoss.pm,v 1.22 2012/12/11 10:08:11 sgadgil Exp $
#
# $sAuthor: $
#
# $Date: 2012/12/11 10:08:11 $
# $Revision: 1.22 $
#
# Provides:                            JBoss
# Author:                              sgadgil
# crAgent Build:                       \$Id: crAgent,v 1.100 201
# crAgent Run Date:                    12 Nov, 2012
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package JBoss;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::Basename;
use File::stat;
use Carp qw(cluck);

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT ||= "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: JBoss.pm,v 1.22 2012/12/11 10:08:11 sgadgil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define JBoss subroutines..
#--------------------------------

sub SplitAddress {

  my ( $sManagementAddress, $refHostname, $refPort )  = @_;
  #-----------------
  # Announce Entry..
  #-----------------

  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  ManagementAddress        [%s]",

  $sMyName, $sManagementAddress );

  my ($sHostname, $sPort) = ();

  # check if the IP is IPv6
  # "begins with a '[', non-greedy match of everything, followed by a ']'
  $sHostname = $sManagementAddress =~ /^\[(.*?)\]/;

  if ($sHostname) {
	  # '$sHostname' is now assumed to be an IPv6 address ...
	  # extract the 'port' which could be after the ']:' for an IPv6 url
	  my @url = split ( /\]:/, $sManagementAddress, 2);

	  # if the ']:' split succeeded, ignore the leading '['.
	  # if it fails, it means that no 'port' has been specified; ignore the
	  # first '[' and the last ']'
	  if ( $url[1] ) {
		  # assume the second part is the 'port' ...
		  $sPort = $url[1];
		  $sHostname = substr ($url[0], 1);
	  } else {
		  $sHostname = substr ($url[0], 1, -1);
	  }
  } else {
	  ($sHostname, $sPort) = split (/:/, $sManagementAddress, 2);
  }

  if ( ! $sHostname ) {
    &logError ( 20002, "Invalid Hostname [%s]", $sHostname );
    return 0; #failed
  }

  ${$refHostname} = "$sHostname";

  # validate Port only if 'requested' and/or 'defined' ...
  if ( ! $refPort ) {
    return 1; # success
  }

  # refPort defined, so asking for 'port' ...
  if ( ! $sPort ) {
	  # address doesn't contain a 'port' value ...
	  ${$refPort} = "";
	  return 1; # success
  }

  # check if numeric ...
  if ( $sPort !~ /^\d+$/ ) {
	  &logError ( 20003, "Invalid Port value [%s]", $sPort );
	  return 0; #failed
  }

  if ( $sPort < 1 || $sPort > 65535 ) {
	  &logError ( 20004, "Invalid Port value [%s]", $sPort );
	  return 0; # failed
  }
  ${$refPort} = "$sPort";

  return 1;
}

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String User,
                              IN String EnvFile,
                              IN String AdminUser,
                              IN String AdminPassword,
                              IN String MonitorProgram,
                              IN Integer SecondLevelMonitor,
                              IN String StartArgs,
                              IN String StopArgs,
                              IN String JBossHome,
                              IN String ServerBaseDir,
                              IN String ServerConfig,
                              IN String BindingSet,
                              IN String JNDIUrl,
                              IN String Version,
                              IN String Component,
                              IN String ManagementAddress,
                              IN String MasterHCAddress,
                              IN String ServerName,
			      IN String HostName,
                              IN String ConfigurationDir );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localUser                     System user credentials used to start JBoss (common)
 $localEnvFile                  Absolute path to environment file to source. (common)
 $localAdminUser                JBoss Administrator's username (v5)
 $localAdminPassword            JBoss Administrator password (v5)
 $localMonitorProgram           Full path to external monitor program (common)
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring (common)
 $localStartArgs                JBoss startup arguments. (v5)
 $localStopArgs                 JBoss stop arguments. (v5)
 $localJBossHome                Absolute path to the JBoss Home Directory. (common)
 $localServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $localServerConfig             The server configuration name to operate upon. (common)
 $localBindingSet               Port Binding Set (v5)
 $localJNDIUrl                  JNDI URL for the Server Configuration
 $localVersion                  Version of the JBoss Application Server. (v6)
 $localComponent                The application component for which the resource is configured. (v6)
 $localManagementAddress        IP Address to which process binds. (v6)
 $localMasterHCAddress          For the slave machines, this is the address of the master machine. (v6)
 $localServerName               For component type 'Server', name of the server. (v6)
 $localHostName			For component type 'Server', 'Domain, host name as specified in host.xml
 $localConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localUser,
       $localEnvFile,
       $localAdminUser,
       $localAdminPassword,
       $localMonitorProgram,
       $localSecondLevelMonitor,
       $localStartArgs,
       $localStopArgs,
       $localJBossHome,
       $localServerBaseDir,
       $localServerConfig,
       $localBindingSet,
       $localJNDIUrl,
       $localVersion,
       $localComponent,
       $localManagementAddress,
       $localMasterHCAddress,
       $localServerName,
       $localHostName,
       $localConfigurationDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20005, "Subroutine <%s> called with:
  EntryPointName           [%s]
  User                     [%s]
  EnvFile                  [%s]
  AdminUser                [%s]
  AdminPassword            [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]
  StartArgs                [%s]
  StopArgs                 [%s]
  JBossHome                [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  BindingSet               [%s]
  JNDIUrl                  [%s]
  Version                  [%s]
  Component                [%s]
  ManagementAddress        [%s]
  MasterHCAddress          [%s]
  ServerName               [%s]
  HostName		   [%s]
  ConfigurationDir         [%s]",

  $sMyName,    $sEntryPointName,     $localUser,     $localEnvFile,
  $localAdminUser,    $localAdminPassword,    $localMonitorProgram,
  $localSecondLevelMonitor,    $localStartArgs,     $localStopArgs,
  $localJBossHome,     $localServerBaseDir,     $localServerConfig,
  $localBindingSet, $localJNDIUrl, $localVersion,  $localComponent,
  $localManagementAddress, $localMasterHCAddress, $localServerName,
  $localHostName, $localConfigurationDir );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName    User    EnvFile     AdminUser     AdminPassword
       MonitorProgram SecondLevelMonitor  StartArgs  StopArgs  JBossHome
       ServerBaseDir ServerConfig BindingSet JNDIUrl  Version  Component
       ManagementAddress MasterHCAddress ServerName ConfigurationDir );

  @hsArgList { @lsArgs } = @_; # Assign..

  my @lsOnlineArgs  = ();
  my @lsOfflineArgs = ();
  my @lsMonitorArgs = ();
  my @lsCleanArgs   = ();
  my @lsImfRegisterArgs = ();
  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  if ( $localVersion =~ /^5\.?\d*.*/s ) {
	  @lsOnlineArgs  = qw ( AdminUser AdminPassword User JBossHome ServerBaseDir ServerConfig BindingSet JNDIUrl );
	  @lsOfflineArgs = qw ( AdminUser AdminPassword User JBossHome ServerBaseDir ServerConfig BindingSet JNDIUrl );
	  @lsMonitorArgs = qw ( AdminUser AdminPassword User JBossHome ServerBaseDir ServerConfig BindingSet JNDIUrl );
	  @lsCleanArgs   = qw ( AdminUser AdminPassword User JBossHome ServerBaseDir ServerConfig BindingSet JNDIUrl );
	  @lsImfRegisterArgs = qw ( AdminUser AdminPassword User JBossHome ServerBaseDir ServerConfig BindingSet JNDIUrl );

  } elsif ( ($localVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {
	  @lsOnlineArgs  = qw ( User Component JBossHome ManagementAddress ServerBaseDir ServerConfig );
	  @lsOfflineArgs = qw ( User Component JBossHome ManagementAddress ServerBaseDir ServerConfig );
	  @lsMonitorArgs = qw ( User Component JBossHome ManagementAddress ServerBaseDir ServerConfig );
	  @lsCleanArgs   = qw ( User Component JBossHome ManagementAddress ServerBaseDir ServerConfig );
	  @lsImfRegisterArgs = qw ( User Component JBossHome ManagementAddress ServerBaseDir ServerConfig );
  }

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
                    'imf_register'  => \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20006,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #------------------------
  # Validate parameter User..
  #------------------------
  if ( ! defined getpwnam $localUser ) {
     &logError ( 20007,
      "<%s> User [%s] is not a valid UNIX login",
     $sMyName, $localUser );
     $iReturnCode = 0; # Failure..
  }

  if ( $localVersion =~ /^5\.?\d*.*/s ) {

	  if ( 1 != ArgsValidv5 (
		  $sEntryPointName,
		  $localEnvFile,
		  $localAdminUser,
		  $localAdminPassword,
		  $localMonitorProgram,
		  $localStartArgs,
		  $localStopArgs,
		  $localJBossHome,
		  $localServerBaseDir,
		  $localServerConfig,
		  $localBindingSet,
		  $localJNDIUrl,
		  $localConfigurationDir ) ) {

		$iReturnCode = 0; # Failure..
	}

  } elsif ( ($localVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {

	  if ( 1 != ArgsValidv6 (
		  $sEntryPointName,
		  $localEnvFile,
		  $localMonitorProgram,
		  $localJBossHome,
		  $localServerBaseDir,
		  $localServerConfig,
		  $localComponent,
		  $localManagementAddress,
		  $localMasterHCAddress,
		  $localServerName,
		  $localHostName,
		  $localConfigurationDir ) ) {

		$iReturnCode = 0; # Failure..
	}
  }

  $iReturnCode ||
  &logWarn ( 20008,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

sub ArgsValidv5 {
  my ( $sEntryPointName,
       $localEnvFile,
       $localAdminUser,
       $localAdminPassword,
       $localMonitorProgram,
       $localStartArgs,
       $localStopArgs,
       $localJBossHome,
       $localServerBaseDir,
       $localServerConfig,
       $localBindingSet,
       $localJNDIUrl,
       $localConfigurationDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20009, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  AdminUser                [%s]
  AdminPassword            [%s]
  MonitorProgram           [%s]
  StartArgs                [%s]
  StopArgs                 [%s]
  JBossHome                [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  BindingSet               [%s]
  JNDIUrl                  [%s]
  ConfigurationDir         [%s]",

  $sMyName, $sEntryPointName, $localEnvFile, $localAdminUser, $localAdminPassword,
  $localMonitorProgram, $localStartArgs, $localStopArgs, $localJBossHome,
  $localServerBaseDir, $localServerConfig, $localBindingSet, $localJNDIUrl,
  $localConfigurationDir );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

   #---------------------------------------
   # Validation for online entry point(s)..
   #---------------------------------------
   if ( $sEntryPointName eq "online" ||
	   $sEntryPointName eq "offline" ) {
	   #-------------------------------
	   # Validate parameter JBossHome..
	   #-------------------------------
	   if ( ! -d $localJBossHome ) {
		   &logError ( 20010,
			   "<%s> JBossHome [%s] directory does not exist",
			   $sMyName, $localJBossHome );
		   $iReturnCode = 0; # Failure..
	   }

	   #-----------------------------------
	   # Validate parameter ServerBaseDir..
	   #-----------------------------------
	   if ( ! -d $localServerBaseDir ) {
		   &logError ( 20011,
			   "<%s> ServerBaseDir [%s] directory does not exist",
			   $sMyName, $localServerBaseDir );
		   $iReturnCode = 0; # Failure..
	   }

	   #---------------------------------------------------------------------
	   # Validate parameter MonitorProgram only during online entry point ...
	   #---------------------------------------------------------------------
	   if ( $sEntryPointName eq "online" ) {
		   if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
			   ! -x $1 ) {
			   &logError ( 20012,
				   "<%s> MonitorProgram [%s] is not executable",
				   $sMyName, $1 );
			   $iReturnCode = 0; # Failure..
		   }
	   }
   }

   $iReturnCode ||
  &logWarn ( 20013,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;

}

sub ArgsValidv6 {
  my ( $sEntryPointName,
       $localEnvFile,
       $localMonitorProgram,
       $localJBossHome,
       $localServerBaseDir,
       $localServerConfig,
       $localComponent,
       $localManagementAddress,
       $localMasterHCAddress,
       $localServerName,
       $localHostName,
       $localConfigurationDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20014, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  JBossHome                [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component                [%s]
  ManagementAddress        [%s]
  MasterHCAddress          [%s]
  ServerName               [%s]
  HostName		   [%s]
  ConfigurationDir         [%s]",

  $sMyName,    $sEntryPointName,   $localEnvFile,   $localMonitorProgram,
  $localJBossHome,     $localServerBaseDir,     $localServerConfig,
  $localComponent, $localManagementAddress, $localMasterHCAddress,
  $localServerName, $localHostName, $localConfigurationDir );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  if ( ( $localComponent ne "Standalone" ) && ( $localComponent ne "Domain" ) && ( $localComponent ne "Server" ) ) {
	&logError ( 20015,
	  "Unsupported Component [%s]. The agent only supports 'Standalone', 'Server' and 'Domain' components of JBoss Application Server.",
	  $localComponent );
	  $iReturnCode = 0;
  }
 if ( ( $localComponent eq "Server" ) && ( $localServerName !~ /\S/o ) ) {
        &logError ( 20177, "ServerName must be specified for Server Component" );
	$iReturnCode = 0;
  }

 if ( ( $localComponent ne "Standalone" ) && ( $localHostName !~ /\S/o ) ) {
	&logError ( 20178, "HostName must be specified for Domain Mode support"	);
	$iReturnCode = 0;
  }

  # for version 6, serverconfig is configuration dir + filename
  my $localServerConfigFile = $localConfigurationDir . "/" . $localServerConfig;

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ||
	$sEntryPointName eq "offline" ) {
    #-------------------------------
    # Validate parameter JBossHome..
    #-------------------------------
    if ( ! -d $localJBossHome ) {
      &logError ( 20016,
      "<%s> JBossHome [%s] directory does not exist",
      $sMyName, $localJBossHome );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------------
    # Validate parameter ServerBaseDir..
    #-----------------------------------
    if ( ! -d $localServerBaseDir ) {
      &logError ( 20017,
      "<%s> ServerBaseDir [%s] directory does not exist",
      $sMyName, $localServerBaseDir );
      $iReturnCode = 0; # Failure..
    }

    #----------------------------------
    # Validate parameter ServerConfig..
    #----------------------------------
    if ( ! -e $localServerConfigFile ) {
      &logError ( 20018,
      "<%s> ServerConfig [%s] file does not exist",
      $sMyName, $localServerConfigFile );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------------------
    # Validate parameter ConfigurationDir..
    #--------------------------------------
    if ( ! -d $localConfigurationDir ) {
      &logError ( 20019,
      "<%s> ConfigurationDir [%s] directory does not exist",
      $sMyName, $localConfigurationDir );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------------------------------------------
    # Validate parameter MonitorProgram only during online entry point ...
    #---------------------------------------------------------------------
    if ( $sEntryPointName eq "online" ) {
      if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
           ! -x $1 ) {
        &logError ( 20020,
        "<%s> MonitorProgram [%s] is not executable",
        $sMyName, $1 );
        $iReturnCode = 0; # Failure..
      }
    }
  }

  $iReturnCode ||
  &logWarn ( 20021,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String ConfigurationDir,
                         IN String EnvFile,
                         IN String JBossHome,
                         IN String ManagementAddress,
                         IN Integer SecondLevelMonitor,
                         IN String ServerBaseDir,
                         IN String ServerConfig,
                         IN String User,
                         IN String EnableCmdOutLogging,
                         IN String ServiceName,
                         IN String UseSystemD );

 Description : Start logic needed for online entry pt

 Parameters :

 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring (common)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $sUser                     System user credentials used to start JBoss (common)
 $bEnableCmdOutLogging      Enables Command output logging
 $ServiceName               Name of the service to start, stop and check status of JBoss
 $bUseSystemD               When enabled, the reource comes online in system.slice

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigurationDir,
       $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $iSecondLevelMonitor,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sBindingSet,
       $sJNDIUrl,
       $sStartArgs,
       $sVersion,
       $sServerName,
       $sHostName,
       $bEnableCmdOutLogging,
       $sServiceName,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20022, "Subroutine <%s> called with:
  ConfigurationDir         [%s]
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  SecondLevelMonitor       [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  BindingSet               [%s]
  JNDIUrl                  [%s]
  StartArgs                [%s]
  Version                  [%s]
  ServerName		   [%s]
  HostName		   [%s]
  EnableCmdOutLogging      [%s]
  ServiceName              [%s]
  UseSystemD               [%s]",

  $sMyName,     $sConfigurationDir,     $sEnvFile,     $sJBossHome,
  $sManagementAddress,    $iSecondLevelMonitor,    $sServerBaseDir,
  $sServerConfig, $sComponent, $sUser, $sAdminPassword, $sAdminUser, $sBindingSet,
  $sJNDIUrl, $sStartArgs, $sVersion, $sServerName, $sHostName, $bEnableCmdOutLogging, $sServiceName, $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  
  #-----------------------------------------
  # Set the file name to save command output
  #-----------------------------------------
  my $sResourceType   = $ENV{'VCS_LOG_AGENT_NAME'};
  my $sResourceName   = $ENV{'VCS_LOG_RESOURCE_NAME'};
  my $sEntryPointName = $ENV{'VCS_LOG_SCRIPT_NAME'};

  my $sOutputFilename = "/dev/null";

  if ($bEnableCmdOutLogging) {
    $sOutputFilename = "/tmp/.VRTS$sResourceType/$sResourceName"."_$sEntryPointName.out";
    if (-f "$sOutputFilename") {
      if (0 == unlink ("$sOutputFilename") ) {
        &logError ( 20023,
          "<%s> Unable to delete output file [%s]",
  	$sMyName, $sOutputFilename );
  	 return 0;
       }
    }
  }
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting JBoss processes..
  #-----------------------------------------
  &logTrace ( 20024,
  "<%s> Killing JBoss processes for this instance", $sMyName );

  if ( JBoss::KillJBossProcesses ( $sConfigurationDir,
                                   $sJBossHome,
                                   $sManagementAddress,
                                   $sServerBaseDir,
                                   $sServerConfig,
				   $sComponent,
                                   $sUser,
			           $sVersion,
                                   $sServerName,
                                   $bUseSystemD ) ) {
    &logTrace ( 20025,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20026,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 );
  $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd();

  # we want the IP, don't care about the port ...
  my ( $sManagementIP, $sManagementPort ) = ();
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  if ( $sVersion =~ /^5\.?\d*.*/s ) {
	# Tokenize Url into protocol, ip and port
	my $sTmp = $sJNDIUrl;
	my $sProtocol = substr ($sTmp, 0, 6);
	$sTmp = substr ($sTmp, 6);
	my @result = split (/:/, $sTmp);
	my $sIpAddress = $result[0];
	my $sPort = $result[1];

	$sCommand = "$sJBossHome/bin/run.sh";
	$sCommandArgs = "-c $sServerConfig -Djboss.server.base.dir=$sServerBaseDir -Djboss.server.base.url=file:$sServerBaseDir -Djboss.service.binding.set=$sBindingSet -Djava.rmi.server.hostname=$sIpAddress -Djboss.bind.address=$sIpAddress $sStartArgs";

  } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {

	  if ( ! SplitAddress ( $sManagementAddress, \$sManagementIP, \$sManagementPort ) ) {
		  &logError ( 20027, "<%s> Invalid Hostname/IP address specified in the variable ManagementAddress [%s]",
			  $sMyName, $sManagementAddress );
		  return 0;
	  }
	  if ($sComponent eq "Standalone"){
	  	  $sCommand = "$sJBossHome/bin/standalone.sh";
		  $sCommandArgs = "-Djboss.server.config.dir=$sConfigurationDir -Djboss.server.base.dir=$sServerBaseDir -Djboss.bind.address.management=$sManagementIP -c $sServerConfig";
	  } elsif ($sComponent eq "Domain"){
		$sCommand = "$sJBossHome/bin/domain.sh";
		$sCommandArgs = "-Djboss.server.config.dir=$sConfigurationDir -Djboss.server.base.dir=$sServerBaseDir -Djboss.bind.address.management=$sManagementIP -c $sServerConfig";
	  } else {
		 $sCommand = "$sJBossHome/bin/jboss-cli.sh";
		 $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort command=/host=$sHostName/server-config=$sServerName:start";
	  }

  }

  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20028,
  "<%s> Starting JBoss instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20029,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $bUseSystemD && !($sComponent eq "Server")) {
        my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
          if ( $sServiceName !~ /\S/ ) {
            $sServiceName = "wildfly";
          }
          if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "start" )){
              &logTrace ( 20143,
              "<%s> RunSystemd was successful",
              $sMyName );
          }
  }
  elsif ( $iReturnCode =
       Sys::RunWithEnvCmdInBG ( $sEnvFile,
                               \$bSourceUserEnv,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $sUser,
                               $sOutputFilename,
                               \$bTimedOut,
                               $iTimeout,
                               dirname($sCommand) ) ) {
    &logTrace ( 20030,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20031,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         JBoss::GetProcState ( \@liPids,
                               $sConfigurationDir,
                               $sJBossHome,
                               $sManagementAddress,
                               $sServerBaseDir,
                               $sServerConfig,
			       $sComponent,
                               $sUser,
			       $sVersion,
			       $sServerName,
			       $bUseSystemD,
                               1 ) && (           # Running FFLM..
         $iSecondLevelMonitor  ?                              # If enabled..
         VCSDef_MonitorExit::ONLINE ==                        # Online after..
         JBoss::SecondLevelMonitor ( $sEnvFile,   # Running SLM..
                                     $sJBossHome,
                                     $sManagementAddress,
				     $sComponent,
                                     $sUser,
                                     $sAdminPassword,
                                     $sAdminUser,
                                     $sJNDIUrl,
                                     $sServerConfig,
                                     $sVersion,
				     $sServerName,
				     $sHostName,
                                     $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20032,
      "<%s> Confirmed start-up of this instance of JBoss server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20033,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String ConfigurationDir,
                        IN String EnvFile,
                        IN String JBossHome,
                        IN String ManagementAddress,
                        IN String ServerBaseDir,
                        IN String ServerConfig,
                        IN String User, 
                        IN String EnableCmdOutLogging,
                        IN String ServiceName,
                        IN String UseSystemD );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $sUser                     System user credentials used to start JBoss (common)
 $EnableCmdOutLogging       Enables Command output logging
 $ServiceName              Name of the service to start, stop and check status of JBoss.
 $bUseSystemD              When enabled, the reource comes online in system.slice
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigurationDir,
       $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sJNDIUrl,
       $sStopArgs,
       $sVersion,
       $sServerName,
       $sHostName,
       $bEnableCmdOutLogging,
       $sServiceName,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20034, "Subroutine <%s> called with:
  ConfigurationDir         [%s]
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  JNDIUrl                  [%s]
  StopArgs                 [%s]
  Version                  [%s]
  ServerName		   [%s]
  HostName		   [%s]
  EnableCmdOutLogging      [%s]
  ServiceName              [%s]
  UseSystemD               [%s]",

  $sMyName,     $sConfigurationDir,     $sEnvFile,     $sJBossHome,
  $sManagementAddress, $sServerBaseDir, $sServerConfig, $sComponent, $sUser,
  $sAdminPassword, $sAdminUser, $sJNDIUrl, $sStopArgs, $sVersion, $sServerName, 
  $sHostName, $bEnableCmdOutLogging, $sServiceName, $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..
  
  #-----------------------------------------
  # Set the file name to save command output
  #-----------------------------------------
  my $sResourceType   = $ENV{'VCS_LOG_AGENT_NAME'};
  my $sResourceName   = $ENV{'VCS_LOG_RESOURCE_NAME'};
  my $sEntryPointName = $ENV{'VCS_LOG_SCRIPT_NAME'};
  my $sOutputFilename    = ""; 
  if ($bEnableCmdOutLogging ){ 
    if (-f "$sOutputFilename") {
      my $sOutputFilename    = "/tmp/.VRTS$sResourceType/$sResourceName"."_$sEntryPointName.out";
      if (0 == unlink ("$sOutputFilename") ) {
        &logError ( 20035,
              "<%s> Unable to delete output file [%s]",
              $sMyName, $sOutputFilename );
        return 0;
      }
    }
  }

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
 
  $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd();

  # we want the IP and the port ...
  my ( $sManagementIP, $sManagementPort ) = ();

  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  if ( $sVersion =~ /^5\.?\d*.*/s ) {

	  $sCommand = "$sJBossHome/bin/shutdown.sh";
	  $sCommandArgs = "-S -u $sAdminUser -p \'$sAdminPassword\' -s $sJNDIUrl $sStopArgs";

  } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {

	  if ( ! SplitAddress ( $sManagementAddress, \$sManagementIP, \$sManagementPort ) ) {
		  &logError ( 20036, "<%s> Invalid Hostname/IP address specified in the variable ManagementAddress [%s]",
			  $sMyName, $sManagementAddress );
		  return 0;
	  }
	  if (! $sManagementPort ) {
		  $sManagementPort = "9999";
	  }
	  $sCommand = "$sJBossHome/bin/jboss-cli.sh";
	  if ( $sComponent eq "Standalone" ) {
	  	$sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort --command=:shutdown";
	  }
	  elsif ( $sComponent eq "Domain") {
		$sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort command=/host=$sHostName:shutdown";
	  }
	  else {
		$sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort command=/host=$sHostName/server-config=$sServerName:stop";
	  }
	  undef $sAdminPassword;
  }

  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20037,
  "<%s> Stopping JBoss with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20038,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $bUseSystemD && !($sComponent eq "Server")) {
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
      if ( $sServiceName !~ /\S/ ){
        $sServiceName = "wildfly";
      }
      if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "stop" )){
          &logTrace ( 20144,
          "<%s> RunSystemd was successful",
          $sMyName );
      }
  }
  elsif ( $iReturnCode =
          JBoss::ExecLocalCmd ( $sUser, 
                                $sEnvFile, 
                                $sCommand, 
                                $sCommandArgs, 
                                \$iExitCode, 
                                $iTimeout, 
                                \@lsCmdout, 
                                $sAdminPassword,
                                $sOutputFilename ) ) {
    &logTrace ( 20039,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20040,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      JBoss::GetProcState ( \@liPids,
                             $sConfigurationDir,
                            $sJBossHome,
                            $sManagementAddress,
                            $sServerBaseDir,
                            $sServerConfig,
			    $sComponent,
                            $sUser,
			    $sVersion,
			    $sServerName,
			    $bUseSystemD,
                            0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20041,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20042,
        "<%s> Confirmed shutdown of JBoss instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting JBoss processes..
  #-----------------------------------------
  &logTrace ( 20043,
  "<%s> Killing JBoss processes for this instance", $sMyName );

  if ( JBoss::KillJBossProcesses ( $sConfigurationDir,
                                   $sJBossHome,
                                   $sManagementAddress,
                                   $sServerBaseDir,
                                   $sServerConfig,
				   $sComponent,
                                   $sUser,
			           $sVersion,
				   $sServerName,
				   $bUseSystemD ) ) {
    &logTrace ( 20044,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20045,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String ConfigurationDir,
                        IN String EnvFile,
                        IN String JBossHome,
                        IN String ManagementAddress,
                        IN String ServerBaseDir,
                        IN String ServerConfig,
                        IN String User, 
                        IN String EnableCmdOutLogging,
                        IN String UseSystemD);
 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $sUser                     System user credentials used to start JBoss (common)
 $EnableCmdOutLogging       Enables Command output logging
 $bUseSystemD               When enabled, the reource comes online in system.slice
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sConfigurationDir,
       $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sJNDIUrl,
       $sStopArgs,
       $sVersion,
       $sServerName,
       $sHostName,
       $bEnableCmdOutLogging,
       $bUseSystemD ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20046, "Subroutine <%s> called with:
  CleanReason              [%s]
  ConfigurationDir         [%s]
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  JNDIUrl                  [%s]
  StopArgs                 [%s]
  Version                  [%s]
  ServerName		   [%s]
  HostName		   [%s]
  EnableCmdOutLogging      [%s]
  UseSystemD               [%s]",

  $sMyName,    $sCleanReason,    $sConfigurationDir,     $sEnvFile,
  $sJBossHome,        $sManagementAddress,         $sServerBaseDir,
  $sServerConfig, $sComponent, $sUser, $sAdminPassword, $sAdminUser, $sJNDIUrl,
  $sStopArgs, $sVersion, $sServerName, $sHostName, $bEnableCmdOutLogging, $bUseSystemD );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  
  #-----------------------------------------
  # Set the file name to save command output
  #-----------------------------------------
  my $sResourceType   = $ENV{'VCS_LOG_AGENT_NAME'};
  my $sResourceName   = $ENV{'VCS_LOG_RESOURCE_NAME'};
  my $sEntryPointName = $ENV{'VCS_LOG_SCRIPT_NAME'};
  
  my $sOutputFilename = "";
  
  if ($bEnableCmdOutLogging) {
    $sOutputFilename    = "/tmp/.VRTS$sResourceType/$sResourceName"."_$sEntryPointName.out";
  
    if (-f "$sOutputFilename") {
	  if (0 == unlink ("$sOutputFilename") ) {
		  &logError ( 20047,
			  "<%s> Unable to delete output file [%s]",
			  $sMyName, $sOutputFilename );
		  return 0;
	  }
    }
  }
  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();

  # we want the IP and the port ...
  my ( $sManagementIP, $sManagementPort ) = ();

  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  if ( $sVersion =~ /^5\.?\d*.*/s ) {

	  $sCommand = "$sJBossHome/bin/shutdown.sh";
	  $sCommandArgs = "-S -u $sAdminUser -p \'$sAdminPassword\' -s $sJNDIUrl $sStopArgs";
 
  } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {
 
	  if ( ! SplitAddress ( $sManagementAddress, \$sManagementIP, \$sManagementPort ) ) {
		  &logError ( 20048, "<%s> Invalid Hostname/IP address specified in the variable ManagementAddress [%s]",
			  $sMyName, $sManagementAddress );
		  return 0;
	  }

	  if (! $sManagementPort ) {
		  $sManagementPort = "9999";
	  }

	  $sCommand = "$sJBossHome/bin/jboss-cli.sh";
	  if ( $sComponent eq "Standalone") {
                $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort --command=:shutdown";
          }
	  elsif ( $sComponent eq "Domain") {
                $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort command=/host=$sHostName:shutdown";
	  }
          else {
                $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort command=/host=$sHostName/server-config=$sServerName:stop ";
          }

	  undef $sAdminPassword;
  }

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20049,
    "<%s> Attempting to stop JBoss with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20050,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( JBoss::ExecLocalCmd ( $sUser, 
                                  $sEnvFile, 
                                  $sCommand, 
                                  $sCommandArgs, 
                                  \$iExitCode, 
                                  $iTimeout, 
                                  \@lsCmdout, 
				  $sAdminPassword,
                                  $sOutputFilename ) ) {
      &logTrace ( 20051, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #-----------------------------------------
  # Whack any persisting JBoss processes..
  #-----------------------------------------
  &logTrace ( 20052,
  "<%s> Killing JBoss processes for this instance", $sMyName );

  if ( JBoss::KillJBossProcesses ( $sConfigurationDir,
                                   $sJBossHome,
                                   $sManagementAddress,
                                   $sServerBaseDir,
                                   $sServerConfig,
				   $sComponent,
				   $sUser,
				   $sVersion,
				   $sServerName,
				   $bUseSystemD ) ) {
    &logTrace ( 20053,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20054,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String ConfigurationDir,
                                   IN String JBossHome,
                                   IN String ManagementAddress,
                                   IN String ServerBaseDir,
                                   IN String ServerConfig,
                                   IN String User,
                                   IN String UseSystemD );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $sUser                     System user credentials used to start JBoss (common)
 $bUseSystemD               When enabled, the reource comes online in system.slice
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigurationDir,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sVersion,
       $sServerName,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20055, "Subroutine <%s> called:
  ConfigurationDir         [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  Version                  [%s]
  ServerName		   [%s]
  UsesystemD		   [%s]",

  $sMyName, $sConfigurationDir,  $sJBossHome,  $sManagementAddress,
  $sServerBaseDir, $sServerConfig, $sComponent, $sUser, $sVersion, $sServerName, $bUseSystemD );

  my @liDummyPids = ();
  my $iReturnCode = JBoss::GetProcState ( \@liDummyPids,
                                           $sConfigurationDir,
                                          $sJBossHome,
                                          $sManagementAddress,
                                          $sServerBaseDir,
                                          $sServerConfig,
					  $sComponent,
                                          $sUser,
					  $sVersion,
					  $sServerName,
					  $bUseSystemD,
                                          1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20056,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20057,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20058,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String ConfigurationDir,
                               IN String JBossHome,
                               IN String ManagementAddress,
                               IN String ServerBaseDir,
                               IN String ServerConfig,
                               IN String User,
                               IN String UseSystemD,
                               IN Integer UseFFLM)

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $sUser                     System user credentials used to start JBoss (common)
 $bUseSystemD               When enabled, the reource comes online in system.slice
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sConfigurationDir,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sVersion,
       $sServerName,
       $bUseSystemD,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20059, "Subroutine <%s> called with:
  OnlinePids               [%s]
  ConfigurationDir         [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  Version                  [%s]
  ServerName		   [%s]
  UseSystemD               [%s]
  UseFFLM                  [%s]",

  $sMyName,   $rliOnlinePids,   $sConfigurationDir,    $sJBossHome,
  $sManagementAddress,  $sServerBaseDir,  $sServerConfig, $sComponent,  $sUser,
  $sVersion, $sServerName, $bUseSystemD, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues, %hiPids,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  
  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20060, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( JBoss::GetSrvrDetails ( $sConfigurationDir, 
                                 $sJBossHome, 
                                 $sManagementAddress, 
                                 $sServerBaseDir, 
                                 $sServerConfig, 
				 $sComponent,
				 $sVersion,
				 $sServerName,
				 $bUseSystemD,
                                 \%hsSrvrDetails ) ) {
      &logTrace ( 20061,
      "<%s> Got details for monitoring JBoss instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20062,
      "<%s> Failed to get details for monitoring JBoss instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nJBoss process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = JBoss::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20063,
      "<%s> JBoss instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20064,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20065,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20066,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20067,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20068, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                  unshift ( @$rliOnlinePids, $iPid );         # Store..
    
                  &logTrace ( 20069,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  $hiProcPids { $sKey } = $iPid;        # Store..
                  
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20070,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20071,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( JBoss::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20072,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20073,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20074,
    "<%s> Found [%s] PIDs [%s] for this instance of JBoss.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20075,
    "<%s> JBoss process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EnvFile,
                          IN String JBossHome,
                          IN String ManagementAddress,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor,
                          IN String User,
                          IN Integer $level_two );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sMonitorProgram           Full path to external monitor program (common)
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring (common)
 $sUser                     System user credentials used to start JBoss (common)
 $level_two                Level Two Monitoring when IMF Enabled
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $sMonitorProgram,
       $iSecondLevelMonitor,
       $sComponent,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sJNDIUrl,
       $sServerConfig,
       $sVersion,
       $sServerName,
       $sHostName,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20076, "Subroutine <%s> called with:
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]
  Component		   [%s]	
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  JNDIUrl                  [%s]
  ServerConfig             [%s]
  Version                  [%s]
  ServerName		   [%s]
  HostName		   [%s]
  LevelTwoMonitor          [%s]",

  $sMyName,    $sEnvFile,     $sJBossHome,     $sManagementAddress,
  $sMonitorProgram, $iSecondLevelMonitor, $sComponent, $sUser, $sAdminPassword,
  $sAdminUser, $sJNDIUrl, $sServerConfig, $sVersion, $sServerName, $sHostName, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( -300 == $level_two ) {
     $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );  
     &logTrace ( 20077, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }

  if ( $level_two >= 1 ) 
  {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         JBoss::DelayedAttributeCheck ( $sEnvFile,
                                        $sJBossHome,
                                        $sManagementAddress,
                                        $sUser,
					$sAdminPassword,
					$sAdminUser,
					$sJNDIUrl,
					$sServerConfig,
				        $sVersion ) ) ) {
      &logTrace ( 20078,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20079,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20080,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + VCSagentFW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         JBoss::SecondLevelMonitor ( $sEnvFile,
                                     $sJBossHome,
                                     $sManagementAddress,
				     $sComponent,
                                     $sUser,
				     $sAdminPassword,
				     $sAdminUser,
				     $sJNDIUrl,
				     $sServerConfig,
				     $sVersion,
				     $sServerName,
				     $sHostName,
                                     $iTillWhen ) ) ) {

      &logTrace ( 20081,
      "<%s> JBoss instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20082,
      "<%s> JBoss instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + VCSagentFW::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         JBoss::ExternalMonitor ( $sUser,
                                  $sEnvFile,
                                  $sExtCmd,
                                  $sExtCmdArgs,
                                  $iTillWhen ) ) ) {
      &logTrace ( 20083,
      "<%s> JBoss instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20084,
      "<%s> JBoss instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EnvFile,
                                        IN String JBossHome,
                                        IN String ManagementAddress,
                                        IN String User );

 Description : Just In Time Validation

 Parameters :

 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sUser                     System user credentials used to start JBoss (common)
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sJNDIUrl,
       $sServerConfig,
       $sVersion ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20085, "Subroutine <%s> called with:
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  JNDIUrl                  [%s]
  ServerConfig             [%s]
  Version                  [%s]",

  $sMyName, $sEnvFile, $sJBossHome, $sManagementAddress, $sUser,
  $sAdminPassword, $sAdminUser, $sJNDIUrl, $sServerConfig, $sVersion );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my ( @lsArgs, @lsReqdArgs) = ();

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  @lsArgs = qw ( EnvFile JBossHome ManagementAddress User AdminPassword AdminUser JNDIUrl ServerConfig Version );

  if ( $sVersion =~ /^5\.?\d*.*/s ) {
	  @lsReqdArgs = qw ( AdminPassword AdminUser JBossHome  JNDIUrl  ServerConfig User Version );
  } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {
	  @lsReqdArgs = qw ( JBossHome ManagementAddress User Version );
  }

  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20086,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #---------------------------
  # Validate parameter User..
  #---------------------------
  if ( ! defined getpwnam $sUser ) {
    &logError ( 20087,
    "<%s> User [%s] is not a valid UNIX login",
    $sMyName, $sUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------------------
  # Validate parameter JBossHome..
  #--------------------------------
  if ( ! -d $sJBossHome ) {
    &logError ( 20088,
    "<%s> JBossHome [%s] directory does not exist",
    $sMyName, $sJBossHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EnvFile,
                                     IN String JBossHome,
                                     IN String ManagementAddress,
                                     IN String User,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEnvFile                  Absolute path to environment file to source. (common)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sUser                     System user credentials used to start JBoss (common)
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEnvFile,
       $sJBossHome,
       $sManagementAddress,
       $sComponent,
       $sUser,
       $sAdminPassword,
       $sAdminUser,
       $sJNDIUrl,
       $sServerConfig,
       $sVersion,
       $sServerName,
       $sHostName,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20089, "Subroutine <%s> called with:
  EnvFile                  [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  Component		   [%s]
  User                     [%s]
  AdminPassword            [%s]
  AdminUser                [%s]
  JNDIUrl                  [%s]
  ServerConfig             [%s]
  Version                  [%s]
  ServerName		   [%s]
  HostName		   [%s]
  TillWhen                 [%s]",

  $sMyName, $sEnvFile,  $sJBossHome,  $sManagementAddress, $sComponent, $sUser,
  $sAdminPassword, $sAdminUser, $sJNDIUrl, $sServerConfig, $sVersion, $sServerName,
  $sHostName, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;
  
  #-----------------------------------------
  # Set the file name to save command output
  #-----------------------------------------
  my $sResourceType   = $ENV{'VCS_LOG_AGENT_NAME'};
  my $sResourceName   = $ENV{'VCS_LOG_RESOURCE_NAME'};
  my $sEntryPointName = $ENV{'VCS_LOG_SCRIPT_NAME'};

  #my $sOutputFilename    = "/tmp/.VRTS$sResourceType/$sResourceName"."_$sEntryPointName.slm.out";
  my $sOutputFilename = "";
  if ( -f "$sOutputFilename" &&
	  ( 0 == unlink ("$sOutputFilename") ) ) {
	  &logError ( 20090,
		  "<%s> Unable to delete output file [%s]",
		  $sMyName, $sOutputFilename );

	  return 0;
  }
  
  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();

  # we want the IP and the port ...
  my ( $sManagementIP, $sManagementPort ) = ();

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  if ( $sVersion =~ /^5\.?\d*.*/s ) {

	  $sCommand = "$sJBossHome/bin/twiddle.sh";
	  $sCommandArgs = "-u $sAdminUser -p \'$sAdminPassword\' -s $sJNDIUrl get jboss.system:type=ServerConfig ServerName";

  } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&
                  ($1 > "5") ) {

	  if ( ! SplitAddress ( $sManagementAddress, \$sManagementIP, \$sManagementPort ) ) {
		  &logError ( 20091, "<%s> Invalid Hostname/IP address specified in the variable ManagementAddress [%s]",
			  $sMyName, $sManagementAddress );
		  return 0;
	  }
	  if (! $sManagementPort ) {
		  $sManagementPort = "9999";
	  }

	  $sCommand = "$sJBossHome/bin/jboss-cli.sh";
	  if ( $sComponent eq "Standalone") {
	  $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort --command=\"read-attribute server-state\"";
	  }
	  elsif ( $sComponent eq "Domain") {
	  $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort --command=/host=$sHostName:read-attribute\\\(name=host-state\\\)";
	  }
	  else {
	  $sCommandArgs = "--connect --controller=$sManagementIP:$sManagementPort --command=/host=$sHostName/server-config=$sServerName:read-attribute\\\(name=status\\\)"
	  }

	  # ensure that the 'AdminPassword' is not defined for version 6.0
	  undef $sAdminPassword;
  }

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20092,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( JBoss::ExecLocalCmd ( $sUser, 
                                $sEnvFile, 
                                $sCommand, 
                                $sCommandArgs, 
                                \$iExitCode, 
                                $iTimeout, 
                                \@lsCmdout, 
                                $sAdminPassword, 
                                $sOutputFilename ) ) {
    &logTrace ( 20093,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    if ( $sComponent eq "Standalone") {
	&logTrace ( 20179, "Application successfully running in Standalone mode" );
    }
    elsif ( $sComponent eq "Domain") {
    	my $sCmdOut = $lsCmdout[2];
	if ($sCmdOut =~ /\"running\"/) {
	  &logTrace ( 20180, "Application successfully running in Domain mode" );
	}
	else {
	  &logError ( 20181, "Level-two monitoring for Domain mode component failed");
	  $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
	}
    }
    elsif ( $sComponent eq "Server") {
        my $sCmdOut = $lsCmdout[2];
        if ($sCmdOut =~ /\"STARTED\"/) {
          &logTrace ( 20182, "Server Successfully running" );
        }
        else {
          &logError ( 20183, "Level-two monitoring for Server failed" );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
    }

  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String User,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sUser                     System user credentials used to start JBoss (common)
 $sEnvFile                  Absolute path to environment file to source. (common)
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20094, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20095,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20096,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20097,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20098,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20099,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  JBoss::ExecLocalCmd ( $sUser, 
                        $sEnvFile, 
                        $sCmd, 
                        $sCmdArgs, 
                        \$iExitCode, 
                        $iTimeout, 
                        \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20100,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20101,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20102, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20103, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20104,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20105,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20106,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20107, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20108, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20109, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20110, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20111,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20112,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20113,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( JBoss::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20114,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20115,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20116,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20117,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20118, "<%s> PID [%s] does not exist",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillJBossProcesses ( IN String ConfigurationDir,
                                      IN String JBossHome,
                                      IN String ManagementAddress,
                                      IN String ServerBaseDir,
                                      IN String ServerConfig,
                                      IN String User,
                                      IN String UseSystemD );

 Description  : Kills processes for this instance of JBoss
 
 Parameters :

 $sConfigurationDir       Absolute path in which the server configuration XMLs are present. (v6)
 $sJBossHome              Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress      IP Address to which process binds. (v6)
 $sServerBaseDir          Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig           The server configuration name to operate upon. (common)
 $sUser                   System user credentials used to start JBoss (common)
 $bUseSystemD             When enabled, the reource comes online in system.slice
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillJBossProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigurationDir,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sUser,
       $sVersion,
       $sServerName,
       $bUseSystemD ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20119, "Subroutine <%s> called with:
  ConfigurationDir         [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  User                     [%s]
  Version                  [%s]
  ServerName		   [%s]
  UseSystemD               [%s]",

  $sMyName, $sConfigurationDir,  $sJBossHome,  $sManagementAddress,
  $sServerBaseDir, $sServerConfig, $sComponent, $sUser, $sVersion, $sServerName, $bUseSystemD );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'JBoss::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    JBoss::GetProcState ( \@liGoners,
                          $sConfigurationDir,
                          $sJBossHome,
                          $sManagementAddress,
                          $sServerBaseDir,
                          $sServerConfig,
			  $sComponent,
                          $sUser,
			  $sVersion,
			  $sServerName,
			  $bUseSystemD,
                          0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20120,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20121,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20122,
    "<%s> Could not locate any process this instance of JBoss",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sHideString ,
       $sOutputFilename ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20123, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20124,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {                             # Get decrypted..
                          s/([^\w-])/\\$1/g || 1          # Escape non-words.. 
                        } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20125, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20126,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20127,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20128,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20129,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           $sOutputFilename,
                                                           dirname ($sCommand),
                                                           $sDeCrPasswd);

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20130,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20131,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20132,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20133,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20134, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20135,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if ( "solaris" eq $sSwitch ) {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20136, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20137,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20138,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String ConfigurationDir,
                                  IN String JBossHome,
                                  IN String ManagementAddress,
                                  IN String ServerBaseDir,
                                  IN String ServerConfig,
                                  IN String UseSystemD,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sConfigurationDir         Absolute path in which the server configuration XMLs are present. (v6)
 $sJBossHome                Absolute path to the JBoss Home Directory. (common)
 $sManagementAddress        IP Address to which process binds. (v6)
 $sServerBaseDir            Absolute path to the JBoss Server Configurations Directory (common)
 $sServerConfig             The server configuration name to operate upon. (common)
 $bUseSystemD              When enabled, the reource comes online in system.slice
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigurationDir,
       $sJBossHome,
       $sManagementAddress,
       $sServerBaseDir,
       $sServerConfig,
       $sComponent,
       $sVersion,
       $sServerName,
       $bUseSystemD,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20139, "Subroutine <%s> called with:
  ConfigurationDir         [%s]
  JBossHome                [%s]
  ManagementAddress        [%s]
  ServerBaseDir            [%s]
  ServerConfig             [%s]
  Component		   [%s]
  Version                  [%s]
  ServerName		   [%s]
  UseSystemD               [%s]
  SrvrDetails              [%s]",

  $sMyName, $sConfigurationDir,  $sJBossHome,  $sManagementAddress,
  $sServerBaseDir, $sServerConfig, $sComponent, $sVersion, $sServerName, $bUseSystemD, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
 
  #cluck "this is how we got here\n";
  
  $bUseSystemD = $bUseSystemD && SystemD::DetectSystemd();
  my @lsSrchPatterns = ();
  my ( $sManagementIP, $sManagementPort ) = ();
  # we want the IP, don't care about the port ...
  if ( ! SplitAddress ( $sManagementAddress, \$sManagementIP, \$sManagementPort ) ) {
	  &logError ( 20140, "<%s> Invalid Hostname/IP address specified in the variable ManagementAddress [%s]",
		  $sMyName, $sManagementAddress );
	  return 0;
  }

  if( $sComponent eq "Server" ) {
    @lsSrchPatterns = ( qr!(java -D\[Server:$sServerName\].*-server.* -Djboss.bind.address(.management)?=$sManagementIP\b.*-Djboss.home.dir=$sJBossHome\b.* -Djboss.server.data.dir=$sJBossHome/domain/servers/$sServerName/data\b)|(java -D\[Server:$sServerName\].*-server.* -Djboss.bind.address.management=$sManagementIP\b.*-Djboss.home.dir=$sJBossHome\b.*-Djboss.server.base.dir=$sServerBaseDir\b.*-Djboss.server.config.dir=$sConfigurationDir\b)! );
  } elsif( $bUseSystemD ) {
        if ($sComponent eq "Standalone"){
          @lsSrchPatterns = ( qr!java -D\[Standalone\] -server.*-Djboss.home.dir=$sJBossHome\b.*-Djboss.server.base.dir=$sServerBaseDir\b.*-c $sServerConfig\b.*-b $sManagementIP\b! );
        } elsif ($sComponent eq "Domain") {
          @lsSrchPatterns = ( qr!java -D\[Process Controller\] -server.*$sJBossHome\b.*-c.*$sServerConfig\b.*-b.*$sManagementIP\b! );
          }
      }
      else {
        if ( $sVersion =~ /^5\.?\d*.*/s ) {

          @lsSrchPatterns = ( qr!java\b.*-Dprogram.name=run.sh\b.*-classpath\s+$sJBossHome/bin/run.jar.*org.jboss.Main -c $sServerConfig\b.*-Djboss.server.base.dir=$sServerBaseDir\b! );

        } elsif ( ($sVersion  =~ /^(\d+)\.?\d*.*/s) &&  ($1 > "5") ) {
            if ($sComponent eq "Standalone"){
            @lsSrchPatterns = ( qr!java -D\[Standalone\] -server.*-Djboss.home.dir=$sJBossHome\b.*-Djboss.server.config.dir=$sConfigurationDir\b.*-Djboss.server.base.dir=$sServerBaseDir\b.*-Djboss.bind.address.management=$sManagementIP\b -c $sServerConfig\b! );
            } elsif ($sComponent eq "Domain") {
            @lsSrchPatterns = ( qr!java -D\[Process Controller\] -server.*$sJBossHome\b.*-Djboss.server.config.dir=$sConfigurationDir\b.*-Djboss.server.base.dir=$sServerBaseDir\b.*-Djboss.bind.address.management=$sManagementIP\b -c $sServerConfig\b! );
              }
          }
      }

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  my $sOldMask = umask ( 000 );

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20141, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
    chmod (0777, "/tmp/.VRTS$sResourceType");
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0777 ) ) {
    &logWarn ( 20142, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  umask ( $sOldMask );
  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
